import {
	world,
	system,
	ItemStack,
	EntityInventoryComponent,
	Vector3,
	Block,
	Direction,
	Dimension,
	EntityHitInformation,
	BlockHitInformation,
	Player,
	EffectTypes,
	PlayerInteractWithEntityBeforeEvent,
	PlayerInteractWithEntityBeforeEventSignal,
	ItemStartUseAfterEventSignal,
	ItemStopUseAfterEventSignal,
	ItemStopUseAfterEvent,
	ItemStartUseAfterEvent,
	PlayerLeaveBeforeEventSignal,
	PlayerLeaveBeforeEvent,
	WorldInitializeAfterEventSignal,
	WorldInitializeAfterEvent,
	WorldInitializeBeforeEventSignal,
	WorldInitializeBeforeEvent,
	PlayerInteractWithBlockBeforeEvent,
	PlayerInteractWithBlockBeforeEventSignal,
} from '@minecraft/server';
import ItemHandler from 'items/item.handler';
import CustomPlayer, { TCustomEntity } from 'player/custom-player';
import MeleeWeapon from 'melee/melee-weapon';
import { initializeScoreboard } from 'utils/initialize.scoreboard';
import { HCFComponentManager } from '../modules/item-durability/manager';
import { CANCEL_UI_ITEMS, MELEE_ENTITY_RIGHT_CLICK_BYPASS } from 'config/config';
import { it } from 'node:test';

export interface IItemUseAfterEventParams {
	source: TCustomEntity;
	itemStack: ItemStack;
}

export interface IItemStartUseAfterEventParams {
	source: TCustomEntity;
	itemStack: ItemStack;
	useDuration: number;
}

export interface IProjectilHitAfterEvent {
	dimension: Dimension;
	hitVector: Vector3;
	location: Vector3;
	projectile: TCustomEntity;
	source: TCustomEntity;
	getEntityHit(): EntityHitInformation | undefined;
	getBlockHit(): BlockHitInformation | undefined;
}

export interface IHitBlockAfterEventParams {
	damagingEntity: TCustomEntity;
	hitBlock: Block;
	blockFace: Direction;
}

export interface ITargetHitBlockAfterEventParams {
	source: TCustomEntity;
	hitVector: Vector3;
	previousRedstonePower: number;
	redstonePower: number;
}

export const MELEE_RIGHT_SCOREBOARD: string = 'melee_hit_right';
export const MELEE_LEFT_SCOREBOARD: string = 'melee_hit_left';

export const MELEE_RIGHT_COOLDOWN_SCOREBOARD: string = 'melee_hit_right_cooldown';
export const MELEE_LEFT_COOLDOWN_SCOREBOARD: string = 'melee_hit_left_cooldown';

export const MELEE_DRAW_SCOREBOARD: string = 'draw';

export default class MeleeEventsHandler {
	public static MeleeSet: Set<MeleeWeapon> = new Set();

	private readonly onPlayerInteractWithEntityEvent: PlayerInteractWithEntityBeforeEventSignal = world.beforeEvents.playerInteractWithEntity;
	private readonly onItemStartUseEvent: ItemStartUseAfterEventSignal = world.afterEvents.itemStartUse;
	private readonly onItemStopUseEvent: ItemStopUseAfterEventSignal = world.afterEvents.itemStopUse;
	private readonly onPlayerLeaveEvent: PlayerLeaveBeforeEventSignal = world.beforeEvents.playerLeave;
	private readonly onWorldInitialize: WorldInitializeAfterEventSignal = world.afterEvents.worldInitialize;
	private readonly beforeWorldInitialize: WorldInitializeBeforeEventSignal = world.beforeEvents.worldInitialize;
	private readonly beforeInteractWithBlocks: PlayerInteractWithBlockBeforeEventSignal = world.beforeEvents.playerInteractWithBlock;

	private onHoldingEvent: any = system.runInterval(this.checkHolding.bind(this));

	constructor() {
		this.onPlayerInteractWithEntityEvent.subscribe(this.onPlayerInteractWithEntity.bind(this));
		this.onItemStartUseEvent.subscribe(this.onItemStartUse.bind(this));
		this.onItemStopUseEvent.subscribe(this.onItemStopUse.bind(this));
		this.onPlayerLeaveEvent.subscribe(this.onPlayerLeave.bind(this));
		this.onWorldInitialize.subscribe(this.onWorldInitializeMethod.bind(this));
		this.beforeWorldInitialize.subscribe(this.beforeWorldInitializeMethod.bind(this));
		this.beforeInteractWithBlocks.subscribe(this.onPlayerInteractWith.bind(this));
	}

	private beforeWorldInitializeMethod(data: WorldInitializeBeforeEvent): void {
		HCFComponentManager.loadAll(data);
	}

	private onWorldInitializeMethod(): void {
		initializeScoreboard(MELEE_RIGHT_SCOREBOARD);
		initializeScoreboard(MELEE_LEFT_SCOREBOARD);
		initializeScoreboard(MELEE_RIGHT_COOLDOWN_SCOREBOARD);
		initializeScoreboard(MELEE_LEFT_COOLDOWN_SCOREBOARD);
		initializeScoreboard(MELEE_DRAW_SCOREBOARD);
		MeleeWeapon.importInstances();
	}

	private onPlayerInteractWith(data: PlayerInteractWithBlockBeforeEvent) {
		const player = data.player as CustomPlayer;
		const item = player.getHandItem();
		if (item && CANCEL_UI_ITEMS.some(typeId => item.typeId == typeId) && data.block.typeId != 'minecraft:frame' && data.block.typeId != 'minecraft:glow_frame')
			data.cancel = true;
	}

	private onPlayerLeave(data: PlayerLeaveBeforeEvent): void {
		const player = data.player as CustomPlayer;

		if (!player.isInit) return;

		delete player.weapon;

		system.run(() => {
			player.resetScore(MELEE_RIGHT_SCOREBOARD);
			player.resetScore(MELEE_LEFT_SCOREBOARD);
			player.resetScore(MELEE_RIGHT_COOLDOWN_SCOREBOARD);
			player.resetScore(MELEE_LEFT_COOLDOWN_SCOREBOARD);
			player.resetScore(MELEE_DRAW_SCOREBOARD);
		});
	}

	private onItemStartUse({ source, itemStack }: ItemStartUseAfterEvent): void {
		if (source && !(source instanceof Player)) return;
		if (!itemStack) return;

		const player = source as CustomPlayer;

		if (player.startedHoldingRightClickDate) return;
		if (!player.isInit) return;
		if (!player.isAbleToUse) return;

		for (const weapon of MeleeEventsHandler.MeleeSet) {
			if (!ItemHandler.isWeaponItem(itemStack, weapon)) continue;
			if (!weapon.charges.enabled) continue;

			weapon.onStartHoldingRightClick(player);
		}
	}

	private onItemStopUse({ source, itemStack }: ItemStopUseAfterEvent): void {
		if (source && !(source instanceof Player)) return;
		if (!itemStack) return;

		const player = source as CustomPlayer;

		if (!player.isInit) return;
		if (!player.isAbleToUse) return;
		if (player.startedHoldingRightClickDate === null) return;
		if (!player.weapon) return;

		for (const weapon of MeleeEventsHandler.MeleeSet) {
			if (!ItemHandler.isWeaponItem(itemStack, weapon)) continue;

			weapon.onReleaseHoldingRightClick(player);
		}
	}

	private checkHolding(): void {
		for (const p of world.getPlayers()) {
			const player = p as CustomPlayer;

			// console.warn(player.getRotation().x, player.getRotation().y)

			if (!player.isInit) continue;

			const inventory = player.getComponent('inventory') as EntityInventoryComponent;
			const item: ItemStack = inventory.container.getItem(player.selectedSlotIndex);

			if (player.selectedSlotIndex !== player.holdingWeaponSlot /* || !ItemHandler.isWeaponItemInstance(item) */) {
				player.holdingWeaponSlot = player.selectedSlotIndex;
				player.ableToUseDate.reset();
				player.isAbleToUse = false;
				player.weapon = null;
				player.resetScore(MELEE_RIGHT_SCOREBOARD);
				player.resetScore(MELEE_LEFT_SCOREBOARD);
				player.resetScore(MELEE_RIGHT_COOLDOWN_SCOREBOARD);
				player.resetScore(MELEE_LEFT_COOLDOWN_SCOREBOARD);
				player.resetScore(MELEE_DRAW_SCOREBOARD);
				player.resetRightClickHolding();
				player.chargeIndex = null;

				delete player.weapon;
			}

			if (!ItemHandler.isWeaponItemInstance(item)) continue;

			for (const weapon of MeleeEventsHandler.MeleeSet) {
				if (ItemHandler.isWeaponItem(item, weapon)) {
					if (!player.weapon || weapon.itemTypeId !== player.weapon?.itemTypeId) {
						player.weapon = weapon;
						player.drawWeaponLastSlot = player.selectedSlotIndex;
					}

					player.addEffect(EffectTypes.get('weakness'), 1, {
						showParticles: false,
						amplifier: 255,
					});

					player.weapon.onHolding(player);
					player.isHoldingWeapon = true;
					player.holdingWeaponSlot = player.selectedSlotIndex;
				}
			}
		}
	}

	private onPlayerInteractWithEntity(data: PlayerInteractWithEntityBeforeEvent): void {
		const player = data.player as CustomPlayer;
		if (!player && !player.isValid()) return;

		if (MELEE_ENTITY_RIGHT_CLICK_BYPASS.includes(data.target?.typeId))
			return;
		if (player.isSneaking) return;
		if (!player.weapon) return;
		if (!player.weapon?.rightAttack?.enabled) return;

		data.cancel = true;
	}
}
